--
-- Beispiele fr binre Bume
--

module Btree where

-- die Typdefinition des binren Baumes
data Btree a = Leaf a
	| Fork (Btree a) (Btree a)
	deriving (Show)

-- kovertierung von Liste nach Baum, nicht optimiert
mkBtree' :: [a] -> Btree a
mkBtree' [] = error ""
mkBtree' xs@(x:xs')
	| m == 0 = Leaf (x)
	| otherwise = Fork xt yt
	where
	m = (length xs) `div` 2 
	(xt,yt) = (mkBtree' (x:[]), mkBtree' xs')

-- kostruktion eines Baumes mit minimaler Hhe
mkBtree :: [a] -> Btree a
mkBtree [] = error ""
mkBtree xs
	| (m == 0) = Leaf (unwrap xs)
	| otherwise = Fork (mkBtree xt) (mkBtree yt)
	where
	m = (length xs) `div` 2
	(xt, yt) = splitAt m xs
	unwrap [x] = x

-- gre des Baumes (Anzahl der externen Knoten), entspr. length
size :: Btree a -> Int
size (Leaf x) = 1
size (Fork xt yt) = size xt + size yt


-- Hhe des Baumes, Anzahl der Ebenen
height :: Btree a -> Int
height (Leaf x) = 0
height (Fork xt yt) = 1 + max (height xt) (height yt)

-- Anzahl der internen Knoten
nodes :: Btree a -> Int
nodes (Leaf x) = 0
nodes (Fork xt yt) = 1 + nodes xt + nodes yt

-- erstellt einen neuen Baum, dessen Blattinformation = seiner Tiefe ist
depths :: Btree a -> Btree Int
depths = down 0
	where
	down :: Int -> Btree a -> Btree Int
	down n (Leaf x) = Leaf n
	down n (Fork xt yt) = Fork (down (n+1) xt) (down (n+1) yt)

-- maxBtree Ermitteln des maximalen Wertes im Baum
maxBtree :: (Ord a) => Btree a -> a
maxBtree (Leaf f) = f
maxBtree (Fork xt yt) = max (maxBtree xt) (maxBtree yt)


-- map Funktiont fr Btrees, mit map f . flatten = flatten . mapBtree f
mapBtree :: (a -> b) -> Btree a -> Btree b
mapBtree f (Leaf x) = Leaf (f x)
mapBtree f (Fork xt yt) = Fork (mapBtree f xt) (mapBtree f yt)

-- fold Funktion fr Btrees
foldBtree :: (a -> b) -> (b -> b -> b) -> Btree a -> b
foldBtree f g (Leaf x) = f x
foldBtree f g (Fork xt yt) = g (foldBtree f g xt) (foldBtree f g yt)



-- Funktionen realisiert mit foldBtree

size2 = foldBtree (const 1) (+)

height2 = foldBtree (const 0) maxsubs
		where m `maxsubs` n = 1 + m `max` n
		
maxBtree2i = foldBtree (max 0) (max) 

-- erzeugen einer Instanz der Klasse Ord
maxBtree2 :: (Ord a) => Btree a -> a
maxBtree2 = foldBtree id (max)
		-- where hello m = m 
		

	
flatten :: Btree a -> [a]
flatten (Leaf x) = [x]
flatten (Fork xt yt) = flatten xt ++ flatten yt	
		
-- flatten mit foldBtree
flatten2 = foldBtree (wrap) (++)
		where wrap x = [x]
			
		
maxsubs m n = 1 + m `max` n

-- Beispielbume
t1 = mkBtree ["a","b","c"]
t2 = mkBtree ["a","b","c","d","e","f","g"]
t3 = mkBtree' [1,2,3]
t4 = mkBtree' [1,2,3,4,5,6,7]
t5 = mkBtree [1,2]




