--
-- Beispiele fr binre Halden
--
module Htree where

-- Datenstruktur Htree
data (Ord a) => Htree a = Null
						| Fork a (Htree a) (Htree a)
						deriving(Show)

-- mkHtree:						
-- Aufbau einer binren Halde
mkHtree :: (Ord a) => [a] -> Htree a
mkHtree = heapify . mkTree

-- mkTree:
-- Aufbau eines minimalen Baumes
mkTree :: (Ord a) => [a] -> Htree a
mkTree [] = Null -- Leerer baum
mkTree [x] = Fork x Null Null -- ein neuer Knoten
mkTree (x:xs) = Fork x (mkTree ys) (mkTree zs) 
	where
	m = (length xs) `div` 2
	(ys, zs) = splitAt m xs	

-- sift:
-- Durchsieben der Knoten, eigentliches Sortieren
heapify :: (Ord a) => Htree a -> Htree a
heapify Null = Null
heapify (Fork x xt yt) = sift x (heapify xt) (heapify yt)

sift :: (Ord a) => a -> Htree a -> Htree a -> Htree a
sift x Null Null = Fork x Null Null
sift x (Fork y a b) Null = if x <= y then Fork x (Fork y a b) Null else Fork y (sift x a b) Null
sift x Null (Fork z c d) = if x <= z then Fork x Null (Fork z c d) else Fork z Null (sift x c d)
sift x (Fork y a b) (Fork z c d)
	| x <= (y `min` z) = Fork x (Fork y a b) (Fork z c d)
	| y <= (x `min` z) = Fork y (sift x a b) (Fork z c d)
	| z <= (x `min` y) = Fork z (Fork y a b) (sift x c d)


-- Flatten:
-- Baum in Liste konvertieren
flatten :: (Ord a) => Htree a -> [ a ]
flatten Null = []
flatten (Fork x xt yt)= x:merge (flatten xt) (flatten yt)

-- Merge:
-- 2 Listen in eine dritte, sortierte Liste berfhren
merge :: (Ord a) => [a] -> [a] -> [a]
merge [ ] ys = ys
merge (x:xs) [] = x:xs
merge (x:xs) (y:ys) = if x <= y then x:merge xs (y:ys) else y:merge (x:xs) ys 
